import panndas.nn.functional as F

from .module import Module


class ReLU(Module):
    """Ol' ReLU-iable."""

    def forward(self, xs):
        return F.relu(xs)


class Mish(Module):
    """Applies the Mish function, element-wise.

    For details, see Mish: A Self-Regularized Non-Monotonic
    Neural Activation Function."""

    def forward(self, xs):
        """Applies the Mish function, element-wise."""
        return F.mish(xs)


class Sigmoid(Module):
    """Applies the sigmoid function, element-wise."""

    def forward(self, xs):
        """Applies the sigmoid function, element-wise."""
        return F.sigmoid(xs)


class Tanh(Module):
    """Applies the hyperbolic tangent function, element-wise."""

    def forward(self, xs):
        """Applies the hyperbolic tangent function, element-wise."""
        return F.tanh(xs)


class Softplus(Module):
    """Applies the softplus function, element-wise."""

    def forward(self, xs):
        """Applies the softplus function, element-wise."""
        return F.softplus(xs)


class Softmax(Module):
    """Applies softmax function, column-wise."""

    def forward(self, xs):
        """Applies softmax function, column-wise."""
        return F.softmax(xs)


class Attention(Module):
    """Literally all you need."""

    def __init__(self, queries_df, keys_df, values_df):
        self.w_q = queries_df
        self.w_k = keys_df
        self.w_v = values_df

    def forward(self, xs):
        """Computes queries, keys, and values given inputs."""
        Q = self.w_q @ xs
        K = self.w_k @ xs
        V = self.w_v @ xs

        return Q, K, V


class LinearAttention(Attention):
    """The most basic version of an attention layer."""

    def forward(self, xs):
        """Combines queries, keys, and values linearly."""
        V, Q, K = super().forward(xs)

        A = Q.T @ K
        zs = V @ A

        return zs


class SoftmaxAttention(Attention):
    """The best-known version of an attention layer."""

    def forward(self, xs):
        """Uses a softmax over the sequence dim to select which values to attend to."""
        V, Q, K = super().forward(xs)

        A = F.softmax((Q.T @ K).T).T
        zs = V @ A

        return zs



from .module import Module


class AdditiveSkip(Module):
    """A Module that applies an additive "skip" connection around the provided Module."""

    def __init__(self, block):
        self.block = block

    def forward(self, xs):
        return xs + self.block(xs)

    def show(self):
        return f"[{super().show()}[{self.block.show()}]"


class TransformerEncoderLayer(Module):
    pass



"""Functional interface."""
import math

import pandas as pd


def _pointwise_map(p, f):
    if isinstance(p, pd.Series):
        return p.apply(f)
    else:
        return p.applymap(f)


def _columnwise_map(p, f):
    return p.apply(f, axis=0)


def _rowwise_map(p, f):
    return p.apply(f, axis=1)


def relu(p):
    return _pointwise_map(p, _relu)


def _relu(x):
    return max(x, 0.0)


def softplus(p, beta=1.0, threshold=20.0):
    return _pointwise_map(p, _softplus)


def _softplus(x, beta, threshold):
    if x * beta > threshold:
        return x
    else:
        return 1.0 / beta * math.log(1 + math.exp(beta * x))


def tanh(p):
    return _pointwise_map(p, _tanh)


def _tanh(x):
    return (math.exp(x) - math.exp(-x)) / (math.exp(x) + math.exp(-x))


def mish(p):
    return _pointwise_map(p, _mish)


def _mish(x):
    return softplus(tanh(x))


def sigmoid(p):
    return _pointwise_map(p, _sigmoid)


def _sigmoid(x):
    return 1 / (1 + math.exp(-x))


def softmax(p):
    if isinstance(p, pd.DataFrame):
        return _columnwise_map(p, _softmax)
    else:
        return _softmax(p)


def _softmax(s):
    exps = s.apply(math.exp)
    return exps / exps.sum()


def entropy(*args, **kwargs):
    raise NotImplementedError


def squared_difference(pX, pY):
    """Computes the squared differences between pX and pY."""
    return _pointwise_map(pX - pY)(lambda r: r**2)


def absolute_difference(pX, pY):
    """Computes the absolute differences between pX and pY."""
    return _pointwise_map(pX - pY)(lambda r: abs(r))



from .module import Module


class Sequential(Module):
    """A Module that applies an iterable of Modules sequentially."""

    def __init__(self, modules):
        self.modules = modules

    def forward(self, xs):
        for module in self:
            xs = module(xs)
        return xs

    def show(self):
        return f"{super().show()}{[module.show() for module in self]}"

    def __iter__(self):
        return iter(self.modules)

    def __getitem__(self, key):
        return self.modules[key]



from itertools import combinations

import pandas as pd

from ..nn import Module


class Tokenizer(Module):
    """Tokenizes by spaces and converts to one-hot vectors."""

    def __init__(self, vocab):
        self.vocab = vocab
        self.vocab_ix = pd.Index(vocab)

    def tokenize_one(self, word):
        return pd.Series(self.vocab_ix == word, index=self.vocab_ix)

    def tokenize_many(self, words):
        return pd.DataFrame({word: self.tokenize_one(word) for word in words})

    def tokenize(self, text):
        words = text.split(" ")
        return self.tokenize_many(words)

    def forward(self, xs):
        return self.tokenize(xs)

    @classmethod
    def from_corpus(cls, corpus):
        """Create a Tokenizer by reading a corpus to determine the vocabulary."""
        vocab = vocab_from_corpus(corpus)
        return cls(vocab)


def vocab_from_corpus(corpus):
    """Create a word list from an iterable of texts, where words are split by spaces."""
    return list(set(sum([text.split(" ") for text in corpus], [])))


def text_to_word(text, word):
    """Chop off all parts of a text after the given word."""
    words = text.split(" ")
    index = words.index(word)
    return " ".join(words[: index + 1])


def to_index(vocab):
    """Create a pandas Index out of an iterable of words."""
    return pd.Index(vocab, name="vocab")


def pairs_from(vocab):
    """Create a list of all possible (order-invariant) pairs of words from a vocabulary."""
    return list(combinations(vocab, 2))


def make_pairs_index(vocab):
    """Create a pandas Index for all pairs of words from a vocabulary. See pairs_from."""
    pairs = pairs_from(vocab)
    pairs_ix = pd.Index(f"{fst} {snd}" for (fst, snd) in pairs)
    return pairs_ix


def make_second_order_index(vocab):
    """Create a pandas Index for all pairs and singles of words from a vocabulary."""
    pair_ix = make_pairs_index(vocab)
    second_order_ix = to_index(vocab).append(pair_ix)
    second_order_ix.name = "features"

    return second_order_ix


def pairs_from_text(text, skip=1):
    """Return all pairs of words that occur in a text."""
    words = text.split(" ")
    skip_pairs = zip(words, words[skip:])

    return skip_pairs


def make_second_order_df(vocab):
    """Create an empty DataFrame with vocab in columns and singlets/pairs in rows."""
    vocab_ix = to_index(vocab)
    second_order_ix = make_second_order_index(vocab)
    indexed_vocab = pd.Series(
        vocab, index=vocab_ix, name="vocab"
    )  # vocab as entries and index

    df = pd.DataFrame(index=second_order_ix, columns=indexed_vocab)
    return df


def skips_to_target(words, window):
    """Create the list of paired words before the target within the given window."""
    words, target = words[-window:], words[-1]
    pairs = [[word, target] for word in words[:-1]]

    return pairs



import pandas as pd

from .module import Module


class Linear(Module):
    """A Module that multiplies its inputs by the weights_df and adds the bias_series.

    Input 'tensors'  can be at most 2-D here: feature (rows) and batch/sequence (columns).

    The weights dataframe should have the input feature space as its column index
    and the output feature space as its row index."""

    def __init__(self, weights_df, bias_series=-1):
        self.w = weights_df
        if not isinstance(bias_series, pd.Series):
            bias_series = pd.Series(
                [bias_series] * len(weights_df), index=weights_df.index
            )
        self.b = bias_series

    def forward(self, xs):
        return (self.w @ xs).add(self.b, axis=0)

    def show(self):
        return f"{super().show()}:: {self.w.columns.name} -> {self.w.index.name}"


class Identity(Module):
    """A Module that returns its inputs unaltered."""

    def forward(self, xs):
        return xs



import random

import panndas.nn.functional as F

from .module import Module


class AlphaDropout(Module):
    """A Module that multiplies its inputs by the weights_df and adds the bias_series.

    Input 'tensors'  can be at most 2-D here: feature (rows) and batch/sequence (columns).

    The weights dataframe should have the input feature space as its column index
    and the output feature space as its row index."""

    def __init__(self, p, alpha=0.0):
        self.p = p
        self.alpha = alpha

    def forward(self, xs):
        xs = F._pointwise_apply(lambda x: x if random.random() > self.p else self.alpha)
        return xs

    def show(self):
        return f"{super().show()}:: {self.w.columns.name} -> {self.w.index.name}"


class Dropout(AlphaDropout):
    def __init__(self, p):
        super().__init__(alpha=0.0, p=p)



from . import functional
from .modules import (
    AdditiveSkip,
    AlphaDropout,
    Dropout,
    Identity,
    LayerMaxNorm,
    Linear,
    LinearAttention,
    Mish,
    Module,
    ReLU,
    Sequential,
    Sigmoid,
    Softmax,
    SoftmaxAttention,
    Softplus,
)

__all__ = [
    "functional",
    "Module",
    "Sequential",
    "AdditiveSkip",
    "Identity",
    "Linear",
    "Mish",
    "ReLU",
    "Sigmoid",
    "Softmax",
    "Dropout",
    "AlphaDropout",
    "LayerMaxNorm",
    "SoftmaxAttention",
    "LinearAttention",
    "Softplus",
]



from abc import abstractmethod


class Module(object):
    """An object that is callable via its .forward method."""

    def __call__(self, xs):
        return self.forward(xs)

    @abstractmethod
    def forward(self, xs):
        """Minimally, define this method to define a Module."""
        raise NotImplementedError

    def show(self):
        """Displays the Module."""
        return str(self.__class__.__name__)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['panndas', 'panndas.data', 'panndas.nn', 'panndas.nn.modules']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3,<2.0']

setup_kwargs = {
    'name': 'panndas',
    'version': '0.0.2',
    'description': 'Neural networks in pandas',
    'long_description': '<div align="center">\n  <img src="https://charlesfrye.github.io/img/panndas-logo.jpeg"><br>\n</div>\n\n-----------------\n\n# `panndas`: neural networks in `pandas`\n[![Tests](https://github.com/charlesfrye/panndas/workflows/Tests/badge.svg)](https://github.com/charlesfrye/panndas/actions?workflow=Tests)\n[![Coverage](https://codecov.io/github/charlesfrye/panndas/coverage.svg?branch=main)](https://codecov.io/gh/charlesfrye/panndas)\n[![Read the Docs](https://readthedocs.org/projects/panndas/badge/)](https://panndas.readthedocs.io/)\n[![PyPI Latest Release](https://img.shields.io/pypi/v/panndas.svg)](https://pypi.org/project/panndas/)\n[![License](https://img.shields.io/pypi/l/panndas.svg)](https://github.com/pandas-dev/pandas/blob/main/LICENSE)\n[![Downloads](https://static.pepy.tech/personalized-badge/panndas?period=month&units=international_system&left_color=black&right_color=orange&left_text=PyPI%20downloads%20per%20month)](https://pepy.tech/project/panndas)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## What is it?\n\n**Pandas** are bears endemic to China characterized by a bold black-and-white coat, a rotund body, and a [remarkable similarity](https://wp.technologyreview.com/wp-content/uploads/2019/05/adversarial-10.jpg) to [gibbons](https://www.memphiszoo.org/assets/2510/10_white_cheek_gibbon.jpg).\n\n**`pandas`** is a Python package that provides fast, flexible, and expressive data\nstructures designed to make working with "relational" or "labeled" data both\neasy and intuitive.\n\n**`panndas`** is a neural network library built on top of `pandas` as a joke.\n\n## Main Features\nHere are just a few of the things that `panndas` does well:\n\n  -\n\n## Frequently asked questions\n\n#### How do you pronounce "`panndas`"?\n\nThere are two accepted pronunciations: "pa-`None`-das" and "🖕".\n\n### Why?\n\n> > > We do this not because it is easy, but because it is hard.\n> >\n> > _Wayne Gretsky_\n>\n> _Michael Scott_\n\n_Me_\n\n## Contributing to `panndas` [![Open Source Helpers](https://www.codetriage.com/charlesfrye/panndas/badges/users.svg)](https://www.codetriage.com/charlesfrye/panndas)\n\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are unwelcome.\n\n## License\n[YOLO](http://swansonquotes.com/wp-content/uploads/s05-ep01-permits1-1000x500.jpg)\n',
    'author': 'Charles Frye',
    'author_email': 'cfrye59@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/charlesfrye/panndas',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from .module import Module


class LayerMaxNorm(Module):
    """Normalize across the feature dimension with respect to the infinity norm."""

    def forward(self, xs):
        return xs.divide(xs.abs().max(axis=0), axis=1).fillna(0.0)


class BatchNorm1d(Module):
    """Normalize across the sequence dimension with respect to the 2 norm."""

    def __init__(self, eps=1e-05, gamma=1.0, beta=0.0):
        self.eps = eps
        self.gamma = gamma
        self.beta = beta

    def forward(self, xs):
        mu = xs.mean()
        sigma = (xs.var() + self.eps).sqrt()
        return (xs - mu) / (sigma) * self.gamma + self.beta



from .activation import (
    LinearAttention,
    Mish,
    ReLU,
    Sigmoid,
    Softmax,
    SoftmaxAttention,
    Softplus,
    Tanh,
)
from .batchnorm import BatchNorm1d, LayerMaxNorm
from .container import Sequential
from .dropout import AlphaDropout, Dropout
from .linear import Identity, Linear
from .module import Module
from .transformer import AdditiveSkip

__all__ = [
    "Module",
    "Sequential",
    "Identity",
    "Linear",
    "ReLU",
    "Mish",
    "Sigmoid",
    "Softmax",
    "Softplus",
    "Tanh",
    "Dropout",
    "AlphaDropout",
    "LinearAttention",
    "SoftmaxAttention",
    "AdditiveSkip",
    "BatchNorm1d",
    "LayerMaxNorm",
]
